import re
import threading
from datetime import datetime

from SapphireQuant.Core.Futures import Future, FutureOrder, FutureAsset

from SapphireQuant.Channel.CtpChannel import thosttraderapi as api

from SapphireQuant.Channel.CtpChannel.CtpConvert import CtpConvert
from SapphireQuant.Core.Enum import EnumBuySell, EnumOpenClose, EnumOrderSubmitStatus, EnumOrderStatus, EnumOrderPriceType


class CtpTradeChannel(api.CThostFtdcTraderSpi):
    """
    CTP交易接口
    """

    def __init__(self, user_id, password, broker_id, auth_code, app_id='client_QiPlatform_1'):
        self._user_id = user_id
        self._password = password
        self._broker_id = broker_id
        self._app_id = app_id
        self._auth_code = auth_code
        self._trade_api = None
        self._trade_spi = None
        self._front_address = None
        self._flow_path = None
        self.is_connected = False
        self.trading_date = None
        self._event = threading.Event()
        self._event.clear()

        self._bind_map = {}
        self._is_front_connected = False
        self._is_authenticate = False
        self._is_login = False
        self._is_settlement_info_confirm = False
        self._settlement_info = None
        self._asset = None
        self._lst_future_info = []
        self._lst_order = []
        self._lst_trade = []
        self._lst_position = []
        self._lst_position_detail = []

        self._max_order_ref = 0
        self._front_id = 0
        self._session_id = 0
        self._request_id = 0
        api.CThostFtdcTraderSpi.__init__(self)

    @property
    def broker_id(self):
        """

        :return:
        """
        return self._broker_id

    def connect(self, front_address, flow_path=None):
        """
        连接前置
        :param front_address:
        :param flow_path:
        """
        print('开始连接期货交易服务器......')
        # print(threading.currentThread().ident)
        if self.is_connected:
            print('交易服务器已连接......')
            return
        self._front_address = front_address
        self._flow_path = flow_path
        if self._flow_path is not None:
            self._trade_api = api.CThostFtdcTraderApi_CreateFtdcTraderApi(self._flow_path)
        else:
            self._trade_api = api.CThostFtdcTraderApi_CreateFtdcTraderApi()
        self._trade_api.RegisterFront(self._front_address)
        self._trade_api.RegisterSpi(self)  # 注册事件类
        self._trade_api.SubscribePrivateTopic(api.THOST_TERT_QUICK)  # 注册公有流
        self._trade_api.SubscribePublicTopic(api.THOST_TERT_QUICK)  # 注册私有流
        # 连接前置服务器
        front_flag = self._connect_front()
        if not front_flag:
            return False

        # 看穿认证
        authenticate_flag = self._req_authenticate()
        if not authenticate_flag:
            return False

        # 登录
        login_flag = self._req_login()
        if not login_flag:
            return False

        # 确认结算单
        settlement_info = self.query_settlement_info(self.trading_date)
        if settlement_info is None:
            return False
        settlement_info_confirm_flag = self._req_settlement_info_confirm()
        if not settlement_info_confirm_flag:
            return False
        self.is_connected = True
        print('交易句柄初始化完毕......')
        return True

    def _connect_front(self):
        self._is_front_connected = False
        self._trade_api.Init()
        self._event.clear()
        if self._event.wait(5):
            return self._is_front_connected

        print('前置交易服务器连接超时.....')
        return False

    def _req_authenticate(self):
        self._is_authenticate = False
        auth_field = api.CThostFtdcReqAuthenticateField()
        auth_field.BrokerID = self._broker_id
        auth_field.UserID = self._user_id
        auth_field.AppID = self._app_id
        auth_field.AuthCode = self._auth_code
        print("看穿认证:BrokerID:{0},UserID:{1},APPID:{2},认证码:{3}......".format(self._broker_id, self._user_id, self._app_id, self._auth_code))
        self._trade_api.ReqAuthenticate(auth_field, 0)
        self._event.clear()
        if self._event.wait(5):
            return self._is_authenticate

        print('看穿认证超时.....')
        return False

    def _req_login(self):
        self._is_login = False
        login_field = api.CThostFtdcReqUserLoginField()
        login_field.BrokerID = self._broker_id
        login_field.UserID = self._user_id
        login_field.Password = self._password
        login_field.UserProductInfo = "sapphire"
        print('请求登录服务器......')
        self._trade_api.ReqUserLogin(login_field, 0)
        self._event.clear()
        if self._event.wait(5):
            return self._is_login

        print('登录服务器超时.....')
        return False

    def _req_settlement_info_confirm(self):
        print('发送确认结算单指令......')
        p_settlement_info_confirm = api.CThostFtdcSettlementInfoConfirmField()
        p_settlement_info_confirm.BrokerID = self._broker_id
        p_settlement_info_confirm.InvestorID = self._user_id
        rtn = self._trade_api.ReqSettlementInfoConfirm(p_settlement_info_confirm, 0)
        if rtn != 0:
            print("发送确认结算单指令失败:{0}".format(rtn))
            return False
        else:
            print("发送确认结算单指令成功......")
            self._event.clear()
            if self._event.wait(5):
                return self._is_settlement_info_confirm

            print('确认结算单超时.....')
            return False

    def dis_connect(self):
        """
        断开连接
        """
        print('开始断开期货交易服务器......')
        logout_field = api.CThostFtdcUserLogoutField()
        logout_field.BrokerID = self._broker_id
        logout_field.UserID = self._user_id
        print('请求登出服务器......')
        rtn = self._trade_api.ReqUserLogout(logout_field, 0)
        if rtn != 0:
            print("发送登出请求失败:{0}".format(rtn))

    def OnFrontConnected(self):

        """
        连接交易服务器回调函数
        """
        print("连接交易服务器成功回调......")
        self._is_front_connected = True
        self._event.set()

    def OnFrontDisconnected(self, nReason):
        """
        交易前置服务器断开连接响应
        :param nReason:
        """
        print('交易服务器断开:{0}'.format(nReason))

    def OnRspAuthenticate(self, pRspAuthenticateField, pRspInfo, nRequestID, bIsLast):
        """
        看穿认证回调
        :param pRspAuthenticateField:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('看穿认证失败:{0}'.format(pRspInfo.ErrorMsg))
            self._is_authenticate = False
            self._event.set()
            return
        if bIsLast:
            print('看穿认证成功:{0}'.format(pRspInfo.ErrorMsg))
            self._is_authenticate = True
            self._event.set()

    def OnRspUserLogin(self, pRspUserLogin, pRspInfo, nRequestID, bIsLast):
        """
        用户登录回调函数
        :param pRspUserLogin:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('登录失败:{0}'.format(pRspInfo.ErrorMsg))
            self._is_login = False
            self._event.set()
            return
        if bIsLast:
            self._max_order_ref = int(pRspUserLogin.MaxOrderRef)
            self._front_id = pRspUserLogin.FrontID
            self._session_id = pRspUserLogin.SessionID
            self.trading_date = datetime.strptime(pRspUserLogin.TradingDay, '%Y%m%d')
            print('登录交易服务器成功......')
            print('从CTP交易系统获得交易日信息【{0}】'.format(pRspUserLogin.TradingDay))
            print('从CTP交易系统获得MaxOrderRef:{0},FrontId:{1},SessionId:{2}'.format(pRspUserLogin.MaxOrderRef, pRspUserLogin.FrontID, pRspUserLogin.SessionID))
            self._is_login = True
            self._event.set()

    def OnRspUserLogout(self, pUserLogout, pRspInfo, nRequestID, bIsLast):
        """
        客户端登出响应
        :param pUserLogout:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        :return:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('登出失败:{0}'.format(pRspInfo.ErrorMsg))
            return
        if bIsLast:
            self.is_connected = False
            print('登出成功:BrokerID:{0},UserID:{1}'.format(pUserLogout.BrokerID, pUserLogout.UserID))
            print('释放交易句柄......')
            self._trade_api.Release()

    def OnRspQrySettlementInfo(self, pSettlementInfo, pRspInfo, nRequestID, bIsLast):
        """
        请求结算单回调函数
        :param pSettlementInfo:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('请求结算单失败:{0}'.format(pRspInfo.ErrorMsg))
            self._settlement_info = None
            self._event.set()
            return
        if pSettlementInfo is not None and pSettlementInfo.Content is not None:
            self._settlement_info += pSettlementInfo.Content
        if bIsLast:
            print('请求结算单成功......')
            self._event.set()

    def OnRspSettlementInfoConfirm(self, pSettlementInfoConfirm, pRspInfo, nRequestID, bIsLast):
        """
        确认结算单回调
        :param pSettlementInfoConfirm:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('确认结算单失败:{0}'.format(pRspInfo.ErrorMsg))
            self._is_settlement_info_confirm = False
            self._event.set()
            return
        if bIsLast:
            print('确认结算单成功......')
            self._is_settlement_info_confirm = True
            self._event.set()

    def OnRspQryInstrument(self, pInstrument, pRspInfo, nRequestID, bIsLast):
        """
        查询合约回报
        :param pInstrument:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        :return:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('查询合约出错:{0}......'.format(pRspInfo.ErrorMsg))
            self._lst_future_info = None
            self._event.set()
            return
        try:
            future = Future()
            future.id = pInstrument.InstrumentID
            future.name = pInstrument.InstrumentName
            future.exchange_id = pInstrument.ExchangeID
            if pInstrument.ExpireDate is not None and pInstrument.ExpireDate != '':
                future.expire_date = datetime.strptime(pInstrument.ExpireDate, '%Y%m%d')
            if pInstrument.OpenDate is not None and pInstrument.OpenDate != '':
                future.open_date = datetime.strptime(pInstrument.OpenDate, '%Y%m%d')
            future.long_margin_ratio = pInstrument.LongMarginRatio
            future.short_margin_ratio = pInstrument.ShortMarginRatio
            future.product_id = pInstrument.ProductID
            future.volume_multiple = pInstrument.VolumeMultiple
            future.price_tick = pInstrument.PriceTick
            future.product_class = CtpConvert.convert_product_class(pInstrument.ProductClass)
            self._lst_future_info.append(future)
        except Exception as e:
            print('{0}:{1}'.format(pInstrument.InstrumentID, str(e)))
        if bIsLast:
            self._event.set()
            print('查询合约完毕......')

    def OnRspQryTradingAccount(self, pTradingAccount, pRspInfo, nRequestID, bIsLast):
        """
        查询账户资金响应
        :param pTradingAccount:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        :return:
        """
        print('接收到查询账户资金的回报')
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('查询账户资金失败:{0}'.format(pRspInfo.ErrorMsg))
            self._asset = None
            self._event.set()
            return
        if bIsLast:
            print('查询账户资金成功......')
            self._asset = CtpConvert.convert_asset(pTradingAccount)
            self._event.set()

    def OnRspQryOrder(self, pOrder, pRspInfo, nRequestID, bIsLast):
        """
        查询委托回报
        :param pOrder:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        :return:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('查询委托失败:{0}'.format(pRspInfo.ErrorMsg))
            self._lst_order = None
            self._event.set()
            return
        if pOrder is None:
            if bIsLast:
                print('共查询到:0个委托记录......')
                self._event.set()
            return
        order = CtpConvert.convert_order(pOrder)
        if order is None:
            return
        order.guid = self.bind_attached_info(order.local_id)
        self._lst_order.append(order)
        if bIsLast:
            print('共查询到:{0}个委托记录......'.format(len(self._lst_order)))
            self._event.set()

    def OnRspQryTrade(self, pTrade, pRspInfo, nRequestID, bIsLast):
        """
        查询委托回报
        :param pTrade:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        :return:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('查询成交失败:{0}'.format(pRspInfo.ErrorMsg))
            self._lst_trade = None
            self._event.set()
            return
        if pTrade is None:
            if bIsLast:
                print('共查询到:0个成交记录......')
                self._event.set()
                return
            return
        trade = CtpConvert.convert_trade(pTrade, self._front_id, self._session_id)
        if trade is None:
            return
        trade.guid = self.bind_attached_info(trade.local_id)
        self._lst_trade.append(trade)
        if bIsLast:
            print('共查询到:{0}个成交记录......'.format(len(self._lst_trade)))
            self._event.set()

    def OnRspQryInvestorPosition(self, pInvestorPosition, pRspInfo, nRequestID, bIsLast):
        """

        :param pInvestorPosition:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        :return:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('查询持仓失败:{0}'.format(pRspInfo.ErrorMsg))
            self._lst_position = None
            self._event.set()
            return
        try:
            if pInvestorPosition is not None:
                self._lst_position.append(CtpConvert.convert_position(pInvestorPosition))
        except Exception as e:
            print(str(e))

        if bIsLast:
            print('共查询到:{0}个持仓记录......'.format(len(self._lst_position)))
            self._event.set()

    def OnRspQryInvestorPositionDetail(self, pInvestorPositionDetail, pRspInfo, nRequestID, bIsLast):
        """

        :param pInvestorPositionDetail:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        :return:
        """
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('查询持仓明细失败:{0}'.format(pRspInfo.ErrorMsg))
            self._lst_position_detail = None
            self._event.set()
            return

        self._lst_position_detail.append(CtpConvert.convert_position_detail(pInvestorPositionDetail))
        if bIsLast:
            print('共查询到:{0}个持仓明细记录......'.format(len(self._lst_position_detail)))
            self._event.set()

    def OnErrRtnOrderAction(self, pOrderAction, pRspInfo):
        """

        :param pOrderAction:
        :param pRspInfo:
        :return:
        """
        # 这边只处理了撤单的情况, 没有处理修改委托单的情况, ThostFtdcActionFlagType.Delete
        # 如果交易所认为报单错误，用户就会收到OnErrRtnOrderAction。
        print('OnErrRtnOrderAction回报......')
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('OnErrRtnOrderAction Error:{0}@{1} '.format(pRspInfo.ErrorMsg, pOrderAction.ActionDate) +
                  '{0}:{1},'.format(pOrderAction.ActionTime, pOrderAction.StatusMsg) +
                  '{0}'.format(pOrderAction.OrderActionStatus) +
                  '@BrokerID:{0},InvestorID:{1},UserID:{2},'.format(pOrderAction.BrokerID, pOrderAction.InvestorID, pOrderAction.UserID) +
                  'RequestID:{0}OrderActionRef:{1},'.format(pOrderAction.RequestID, pOrderAction.OrderActionRef) +
                  '{0}_{1}_{2},'.format(pOrderAction.FrontID, pOrderAction.SessionID, pOrderAction.OrderRef) +
                  'OrderSysID:{0},{1}.{2},'.format(pOrderAction.OrderSysID, pOrderAction.InstrumentID, pOrderAction.ExchangeID) +
                  'LimitPrice:{0},VolumeChange:{1},ActionFlag:{2}'.format(pOrderAction.LimitPrice, pOrderAction.VolumeChange, pOrderAction.ActionFlag))
            order = CtpConvert.convert_order_action(pOrderAction, pRspInfo)
            order.guid = self.bind_attached_info(order.local_id)
            # OnFutureOrderCancelFailedEvent?.BeginInvoke(order.Guid.ToString(), pRspInfo.ErrorMsg, null, null);

    def OnRspOrderAction(self, pInputOrderAction, pRspInfo, nRequestID, bIsLast):
        """

        :param pInputOrderAction:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        """
        # Thost 收到撤单指令，如果没有通过参数校验，拒绝接受撤单指令。用户就会收到OnRspOrderAction 消息，其中包含了错误编码和错误消息。
        print('OnRspOrderAction回报......')
        if pRspInfo is not None and pRspInfo.ErrorID != 0:
            print('OnRspOrderAction Error:{0}'.format(pRspInfo.ErrorMsg) +
                  '@BrokerID:{0},InvestorID:{1},UserID:{2},'.format(pInputOrderAction.BrokerID, pInputOrderAction.InvestorID, pInputOrderAction.UserID) +
                  'RequestID:{0}'.format(pInputOrderAction.RequestID) +
                  'OrderActionRef:{0},'.format(pInputOrderAction.OrderActionRef) +
                  '{0}_{1}_{2},'.format(pInputOrderAction.FrontID, pInputOrderAction.SessionID, pInputOrderAction.OrderRef) +
                  'OrderSysID:{0},{1}.{2},'.format(pInputOrderAction.OrderSysID, pInputOrderAction.InstrumentID, pInputOrderAction.ExchangeID) +
                  'LimitPrice:{0},VolumeChange:{1},ActionFlag:{2}'.format(pInputOrderAction.LimitPrice, pInputOrderAction.VolumeChange, pInputOrderAction.ActionFlag))
            order = CtpConvert.convert_input_order_action(pInputOrderAction, pRspInfo)
            order.guid = self.bind_attached_info(order.local_id)
            # OnFutureOrderCancelFailedEvent?.BeginInvoke(order.Guid.ToString(), pRspInfo.ErrorMsg, null, null);

    def OnErrRtnOrderInsert(self, pInputOrder, pRspInfo):
        """

        :param pInputOrder:
        :param pRspInfo:
        """
        # 交易所认为报单错误,才触发此回调函数
        print('OnErrRtnOrderInsert回报......')
        order = CtpConvert.convert_input_order(pInputOrder, pRspInfo, self._front_id, self._session_id)
        order.guid = self.bind_attached_info(order.local_id)
        print("OnErrRtnOrderInsert:{0}".format(order.to_string()))
        # OnFutureOrderRejectedEvent?.BeginInvoke(order, null, null);

    def OnRspOrderInsert(self, pInputOrder, pRspInfo, nRequestID, bIsLast):
        """

        :param pInputOrder:
        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        """
        # Thost 收到报单指令，如果没有通过参数校验，拒绝接受报单指令。用户就会收到OnRspOrderInsert 消息，其中包含了错误编码和错误消息。
        # 如果Thost接受了报单指令，用户不会收到OnRspOrderInsert，而会收到OnRtnOrder，
        print('OnRspOrderInsert回报......')
        order = CtpConvert.convert_input_order(pInputOrder, pRspInfo, self._front_id, self._session_id)
        order.guid = self.bind_attached_info(order.local_id)
        print("OnRspOrderInsert:{0}".format(order.to_string()))
        # OnFutureOrderRejectedEvent?.BeginInvoke(order, null, null);

    def OnRtnOrder(self, pOrder):
        """
        委托回报  交易所收到报单后，通过校验。用户会收到 OnRtnOrder。
        :param pOrder:
        :return:
        """
        order = CtpConvert.convert_order(pOrder)
        order.guid = self.bind_attached_info(order.local_id)
        print('OnRtnOrder:{0}'.format(order.to_string()))
        if (order.order_submit_status == EnumOrderSubmitStatus.insert_rejected) or (
                order.order_submit_status == EnumOrderSubmitStatus.cancel_rejected) or (
                order.order_submit_status == EnumOrderSubmitStatus.modify_rejected):
            # OnFutureOrderRejectedEvent?.BeginInvoke(order, null, null)
            return

        if (order.order_submit_status == EnumOrderSubmitStatus.accepted) and (order.order_status == EnumOrderStatus.已撤):
            # OnFutureOrderCanceledEvent?.BeginInvoke(order, null, null);
            return

        # OnFutureOrderEvent?.BeginInvoke(order, null, null);

    def OnRtnTrade(self, pTrade):
        """
        交易所收到报单后，通过校验。用户会收到 OnRtnTrade。
        :param pTrade:
        """
        print('成交回报......')
        trade = CtpConvert.convert_trade(pTrade, self._front_id, self._session_id)
        print('OnRtnTrade:{0}'.format(trade.to_string()))
        trade.guid = self.bind_attached_info(trade.local_id)
        # OnFutureTradeEvent?.BeginInvoke(trade, null, null);

    def OnRspError(self, pRspInfo, nRequestID, bIsLast):
        """

        :param pRspInfo:
        :param nRequestID:
        :param bIsLast:
        """
        print('OnRspError:{0}'.format(pRspInfo.ErrorMsg))

    def send_order(self, order):
        """

        :param order:
        """
        self._max_order_ref += 1
        self._request_id += 1
        id = order.instrument_id.split('.')[0]
        order.local_id = CtpConvert.get_ctp_unique_local_id(self._front_id, self._session_id, self._max_order_ref, id)
        self._bind_map[order.local_id] = order.guid

        if order.direction == EnumBuySell.买入:
            ctp_direction = api.THOST_FTDC_D_Buy
        else:
            ctp_direction = api.THOST_FTDC_D_Sell

        if order.open_close == EnumOpenClose.开仓:
            ctp_open_close = '0'
        else:
            ctp_open_close = '1'

        if order.order_price_type == EnumOrderPriceType.市价:
            ctp_order_price_type = api.THOST_FTDC_OPT_AnyPrice
            ctp_limit_price = 0
        elif order.order_price_type == EnumOrderPriceType.限价:
            ctp_order_price_type = api.THOST_FTDC_OPT_LimitPrice
            ctp_limit_price = order.order_price
        else:
            ctp_order_price_type = api.THOST_FTDC_OPT_AnyPrice
            ctp_limit_price = 0

        order_field = api.CThostFtdcInputOrderField()
        order_field.BrokerID = self._broker_id
        order_field.ExchangeID = order.exchange_id
        order_field.InstrumentID = order.instrument_id
        order_field.UserID = self._user_id
        order_field.InvestorID = self._user_id
        order_field.Direction = ctp_direction
        order_field.LimitPrice = ctp_limit_price
        order_field.VolumeTotalOriginal = order.order_volume
        order_field.OrderPriceType = ctp_order_price_type
        order_field.ContingentCondition = api.THOST_FTDC_CC_Immediately  # 立即发单
        order_field.TimeCondition = api.THOST_FTDC_TC_GFD  # 当日有效
        order_field.VolumeCondition = api.THOST_FTDC_VC_AV  # 任何数量
        order_field.CombHedgeFlag = api.THOST_FTDC_BHF_Speculation  # 投机
        order_field.CombOffsetFlag = ctp_open_close
        order_field.GTDDate = ""
        order_field.OrderRef = str(self._max_order_ref)
        order_field.MinVolume = 0
        order_field.ForceCloseReason = api.THOST_FTDC_FCC_NotForceClose
        order_field.IsAutoSuspend = 0
        rtn = self._trade_api.ReqOrderInsert(order_field, self._request_id)
        if rtn != 0:
            print('报单失败......')
        else:
            print('send_order:{0}'.format(order.to_string()))
        return order

    def send_order_limit_price(self, instrument_id, exchange_id, direction, open_close, price, volume):
        """
        限价发单
        :param instrument_id:
        :param exchange_id:
        :param direction:
        :param open_close:
        :param price:
        :param volume:
        """
        order = FutureOrder()
        order.instrument_id = instrument_id
        order.exchange_id = exchange_id
        order.direction = direction
        order.open_close = open_close
        order.order_price = price
        order.order_volume = volume
        order.order_price_type = EnumOrderPriceType.限价
        self.send_order(order)
        return order

    def send_order_any_price(self, instrument_id, exchange_id, direction, open_close, volume):
        """
        市价发单
        :param instrument_id:
        :param exchange_id:
        :param direction:
        :param open_close:
        :param volume:
        """
        order = FutureOrder()
        order.instrument_id = instrument_id
        order.exchange_id = exchange_id
        order.direction = direction
        order.open_close = open_close
        order.order_price = 0
        order.order_volume = volume
        order.order_price_type = EnumOrderPriceType.市价
        self.send_order(order)
        return order

    def cancel_order(self, order):
        """

        :param order:
        """
        print('cancel_order:{0}'.format(order.to_string()))
        self._request_id += 1
        result = CtpConvert.analysis_ctp_local_id(order.local_id)
        if result['success']:
            input_order_action_field = api.CThostFtdcInputOrderActionField()
            input_order_action_field.ActionFlag = api.THOST_FTDC_AF_Delete
            input_order_action_field.BrokerID = self._broker_id
            input_order_action_field.InvestorID = self._user_id
            input_order_action_field.InstrumentID = order.instrument_id
            input_order_action_field.FrontID = result['front_id']
            input_order_action_field.SessionID = result['session_id']
            input_order_action_field.OrderRef = str(result['order_ref'])
            input_order_action_field.ExchangeID = order.exchange_id
            input_order_action_field.OrderSysID = order.order_sys_id
            rtn = self._trade_api.ReqOrderAction(input_order_action_field, self._request_id)
            if rtn != 0:
                print('发送撤单请求失败:{0}'.format(rtn))
        else:
            print('未找到对应委托:{0}'.format(order.local_id))

    def query_instrument(self, instrument_id=None):
        """
        查询柜台合约信息
        :param instrument_id:
        :return:
        """
        self._request_id += 1
        self._lst_future_info = []
        qry_field = api.CThostFtdcQryInstrumentField()
        if instrument_id is not None:
            qry_field.InstrumentID = instrument_id
        rtn = self._trade_api.ReqQryInstrument(qry_field, self._request_id)
        if rtn != 0:
            print('发送查询合约请求失败:{0}'.format(rtn))
            return None

        print('发送查询合约请求成功......')
        self._event.clear()
        if self._event.wait(30):
            return self._lst_future_info

        print('请求查询合约信息超时......')
        return None

    def query_asset(self):
        """
        查询柜台账户资金
        :return:
        """
        self._request_id += 1
        self._asset = None
        qry_asset_field = api.CThostFtdcQryTradingAccountField()
        qry_asset_field.BrokerID = self._broker_id
        qry_asset_field.InvestorID = self._user_id
        rtn = self._trade_api.ReqQryTradingAccount(qry_asset_field, self._request_id)
        if rtn != 0:
            print('发送查询账户资金请求失败:{0}'.format(rtn))
            return None
        print('发送查询账户资金请求成功......')
        self._event.clear()
        if self._event.wait(5):
            return self._asset

        print('请求查询账户超时......')
        return None

    def query_orders(self):
        """

        :return:
        """
        self._request_id += 1
        self._lst_order = []
        qry_field = api.CThostFtdcQryOrderField()
        qry_field.BrokerID = self._broker_id
        qry_field.InvestorID = self._user_id
        rtn = self._trade_api.ReqQryOrder(qry_field, self._request_id)
        if rtn != 0:
            print('发送查询委托请求失败:{0}'.format(rtn))
            return None
        print('发送查询委托请求成功......')
        self._event.clear()
        if self._event.wait(5):
            return self._lst_order

        print('请求查询委托超时......')
        return None

    def query_trades(self):
        """

        :return:
        """
        self._request_id += 1
        self._lst_trade = []
        qry_field = api.CThostFtdcQryTradeField()
        qry_field.BrokerID = self._broker_id
        qry_field.InvestorID = self._user_id
        rtn = self._trade_api.ReqQryTrade(qry_field, self._request_id)
        if rtn != 0:
            print('发送查询成交请求失败:{0}'.format(rtn))
            return None
        print('发送查询成交请求成功......')
        self._event.clear()
        if self._event.wait(5):
            return self._lst_trade

        print('请求查询成交超时......')
        return None

    def query_positions(self):
        """

        :return:
        """
        self._request_id += 1
        self._lst_position = []
        qry_field = api.CThostFtdcQryInvestorPositionField()
        qry_field.BrokerID = self._broker_id
        qry_field.InvestorID = self._user_id
        rtn = self._trade_api.ReqQryInvestorPosition(qry_field, self._request_id)
        if rtn != 0:
            print('发送查询持仓请求失败:{0}'.format(rtn))
            return None
        print('发送查询持仓请求成功......')
        self._event.clear()
        if self._event.wait(5):
            return self._lst_position

        print('请求查询持仓超时......')
        return None

    def query_position_details(self):
        """
        查询持仓明细
        :return:
        """
        self._request_id += 1
        self._lst_position_detail = []
        qry_field = api.CThostFtdcQryInvestorPositionDetailField()
        qry_field.BrokerID = self._broker_id
        qry_field.InvestorID = self._user_id
        rtn = self._trade_api.ReqQryInvestorPositionDetail(qry_field, self._request_id)
        if rtn != 0:
            print('发送查询持仓明细请求失败:{0}'.format(rtn))
            return None
        print('发送查询持仓明细请求成功......')
        self._event.clear()
        if self._event.wait(5):
            return self._lst_position_detail

        print('请求查询持仓明细超时......')
        return None

    def query_settlement_info(self, trading_date):
        """
        请求结算单
        :param trading_date:
        """
        print('发送请求结算单指令......')
        self._settlement_info = ''
        qry_settlement_info_field = api.CThostFtdcQrySettlementInfoField()
        qry_settlement_info_field.BrokerID = self._broker_id
        qry_settlement_info_field.InvestorID = self._user_id
        qry_settlement_info_field.TradingDay = trading_date.strftime('%Y%m%d')
        rtn = self._trade_api.ReqQrySettlementInfo(qry_settlement_info_field, 0)
        if rtn != 0:
            print("发送请求结算单指令失败:{0}".format(rtn))
            return None

        print("发送请求结算单指令成功......")
        self._event.clear()
        if self._event.wait(5):
            return self._settlement_info

        print("请求结算单超时......")
        return None

    def bind_attached_info(self, local_id):
        """
        绑定附加条件
        :param local_id:
        :return:
        """
        guid = None
        if local_id in self._bind_map.keys():
            guid = self._bind_map[local_id]
        return guid

    def query_asset_by_trading_date(self, trading_date):
        """

        :param trading_date:
        :return:
        """
        print('请求结算单:{0}'.format(trading_date.strftime('%Y%m%d')))
        settlement_info = self.query_settlement_info(trading_date)
        asset = FutureAsset()
        asset.trading_date = trading_date
        patt = r'期初结存 Balance b/f：(\s*?)([-]?[\d]+.[\d]+)'
        match_group = re.findall(patt, settlement_info)
        if match_group is not None and len(match_group) >= 1:
            if len(match_group[0]) >= 2:
                asset.pre_fund_balance = float(match_group[0][1])

        patt = r'期末结存 Balance c/f：(\s*?)([-]?[\d]+.[\d]+)'
        match_group = re.findall(patt, settlement_info)
        if match_group is not None and len(match_group) >= 1:
            if len(match_group[0]) >= 2:
                asset.fund_balance = float(match_group[0][1])

        patt = r'手 续 费 Commission：(\s*?)([-]?[\d]+.[\d]+)'
        match_group = re.findall(patt, settlement_info)
        if match_group is not None and len(match_group) >= 1:
            if len(match_group[0]) >= 2:
                asset.fee = float(match_group[0][1])

        patt = r'出 入 金 Deposit/Withdrawal：(\s*?)([-]?[\d]+.[\d]+)'
        match_group = re.findall(patt, settlement_info)
        if match_group is not None and len(match_group) >= 1:
            if len(match_group[0]) >= 2:
                asset.withdrawal = float(match_group[0][1])

        patt = r'交割手续费 Delivery Fee：(\s*?)([-]?[\d]+.[\d]+)'
        match_group = re.findall(patt, settlement_info)
        if match_group is not None and len(match_group) >= 1:
            if len(match_group[0]) >= 2:
                asset.delivery_fee = float(match_group[0][1])

        patt = r'保证金占用 Margin Occupied：(\s*?)([-]?[\d]+.[\d]+)'
        match_group = re.findall(patt, settlement_info)
        if match_group is not None and len(match_group) >= 1:
            if len(match_group[0]) >= 2:
                asset.margin = float(match_group[0][1])

        patt = r'可用资金 Fund Avail.：(\s*?)([-]?[\d]+.[\d]+)'
        match_group = re.findall(patt, settlement_info)
        if match_group is not None and len(match_group) >= 1:
            if len(match_group[0]) >= 2:
                asset.available_cash = float(match_group[0][1])

        patt = r'平仓盈亏 Realized P/L：(\s*?)([-]?[\d]+.[\d]+)'
        match_group = re.findall(patt, settlement_info)
        if match_group is not None and len(match_group) >= 1:
            if len(match_group[0]) >= 2:
                asset.close_profit = float(match_group[0][1])

        patt = r'持仓盯市盈亏 MTM P/L：(\s*?)([-]?[\d]+.[\d]+)'
        match_group = re.findall(patt, settlement_info)
        if match_group is not None and len(match_group) >= 1:
            if len(match_group[0]) >= 2:
                asset.position_profit = float(match_group[0][1])

        return asset



import datetime
import uuid


class Entity:
    """
    父类
    """
    def __init__(self):
        self.account_id = ''
        self.guid = uuid.uuid1()
        self.product_code = -1
        self.strategy_id = -1
        self.param_id = ''
        self.trading_date = datetime.datetime.today()
        self.update_time = datetime.datetime.today()


"""
BarBlock
"""


class DataBlock:
    """
    数据块，用于加快播放，根据current_index向下播放
    """

    def __init__(self, data_list):
        self.data_list = data_list
        self.current_index = -1

    def get_current_data(self):
        """
        获取当前data
        :return:
        """
        if len(self.data_list) > 0 and self.current_index < len(self.data_list):
            return self.data_list[self.current_index]

        return None



import datetime


class TimeSlice:
    """
    时间区间
    """
    def __init__(self, begin_time=datetime.timedelta(), end_time=datetime.timedelta()):
        self._begin_time = begin_time
        self._end_time = end_time
        self._normal = True

    @property
    def begin_time(self) -> datetime.timedelta:
        """

        :return:
        """
        return self._begin_time

    @begin_time.setter
    def begin_time(self, value):
        self._begin_time = value
        self._normal = (self._begin_time.total_seconds() < self._end_time.total_seconds())

    @property
    def end_time(self) -> datetime.timedelta:
        """

        :return:
        """
        return self._end_time

    @end_time.setter
    def end_time(self, value):
        self._end_time = value
        self._normal = (self._begin_time.total_seconds() < self._end_time.total_seconds())

    @property
    def duration(self) -> datetime.timedelta:
        """

        :return:
        """
        return self._end_time - self._begin_time

    @property
    def normal(self) -> bool:
        """

        :return:
        """
        return self._normal

    def contains(self, sp) -> bool:
        """

        :param sp:
        :return:
        """
        if not self._normal:
            return (self._begin_time.total_seconds() <= sp.total_seconds()) | (sp.total_seconds() <= self._end_time.total_seconds())
        return (self._begin_time.total_seconds() <= sp.total_seconds()) & (sp.total_seconds() <= self._end_time.total_seconds())

    def to_string(self):
        """

        :return:
        """
        return "[" + str(self._begin_time) + "-" + str(self._end_time) + "]"


# time_slice = TimeSlice()
# print(str(time_slice.begin_time))
# print(str(time_slice.end_time))
# time_slice.begin_time = datetime.timedelta(seconds=5)
# time_slice.end_time = datetime.timedelta(seconds=6)
# print(str(time_slice.begin_time))
# print(str(time_slice.end_time))
# print(str(time_slice.duration))
# print(time_slice.to_string())



import datetime

from SapphireQuant.Core.TimeSlice import TimeSlice


class TradingFrameHelper:
    """
    交易框架帮助
    """
    @staticmethod
    def parse_date(str_date: str) -> datetime.datetime:
        """

        :param str_date:
        :return:
        """
        year = int(str_date[0:4])
        month = int(str_date[4:6])
        day = int(str_date[6:])
        date = datetime.datetime(year, month, day)
        return date

    @staticmethod
    def parse_time_slice(str_slice: str) -> []:
        """

        :param str_slice:
        :return:
        """
        all_slice = []
        str_slice_arr = str_slice.split(',')
        count = len(str_slice_arr) / 2
        for i in range(int(count)):
            time_slice = TimeSlice()
            time_slice.begin_time = TradingFrameHelper.str_pares_timedelta(str_slice_arr[i * 2])
            time_slice.end_time = TradingFrameHelper.str_pares_timedelta(str_slice_arr[i * 2 + 1])

            all_slice.append(time_slice)
        return all_slice

    @staticmethod
    def str_pares_timedelta(str_timedelta):
        """

        :param str_timedelta:
        :return:
        """
        str_timedelta_arr = str_timedelta.split(':')
        if len(str_timedelta_arr) == 3:
            hour = int(str_timedelta_arr[0])
            minute = int(str_timedelta_arr[1])
            second = int(str_timedelta_arr[2])
        else:
            hour = int(str_timedelta_arr[0])
            minute = int(str_timedelta_arr[1])
            second = 0

        return datetime.timedelta(hours=hour, minutes=minute, seconds=second)

# str_date = '20190808'
# TradingFrameHelper.parse_date(str_date)



from .BarProvider import BarProvider
from .BarProviderInfo import BarProviderInfo
from .BaseBarHelper import BaseBarHelper
from .BinaryDayBarStream import BinaryDayBarStream
from .BinaryMinBarStream import BinaryMinBarStream
from .BinaryReader import BinaryReader
from .BinaryTickStream import BinaryTickStream
from .BinaryWriter import BinaryWriter
from .DataCheck import DataCheck
from .DateTimeOffsetMap import DateTimeOffsetMap
from .FileHeader import FileHeader
from .QiDataController import QiDataController
from .QiDataDirectory import QiDataDirectory


__all__ = ['BarProvider',
           'BarProviderInfo',
           'BaseBarHelper',
           'BinaryDayBarStream',
           'BinaryMinBarStream',
           'BinaryReader',
           'BinaryTickStream',
           'BinaryWriter',
           'DateTimeOffsetMap',
           'FileHeader',
           'QiDataController',
           'QiDataDirectory',
           ]



class CtpBrokerCollection:
    """
    期货Broker集合
    """
    def __init__(self):
        self.default_broker_id = ''
        self.brokers = []

    def add(self, broker):
        """
        添加broker
        :param broker:
        :return:
        """
        for data in self.brokers:
            if data.id == broker.id:
                return

        self.brokers.append(broker)

    def get_broker(self, broker_id):
        """
        获取broker
        :param broker_id:
        :return:
        """
        for broker in self.brokers:
            if broker.id == broker_id:
                return broker

    def get_servers(self, broker_id, group_name, network_type, server_type):
        """
        获取服务器组
        :param broker_id:
        :param group_name:
        :param network_type:
        :param server_type:
        :return:
        """
        broker = self.get_broker(broker_id)
        if broker is not None:
            return broker.get_servers(group_name, network_type, server_type)

        return []

    def get_servers_by_network_server_type(self, network_type, server_type):
        """
        获取服务器组
        :param network_type:
        :param server_type:
        :return:
        """
        broker = self.get_broker(self.default_broker_id)
        if broker is not None:
            return broker.get_servers(network_type, server_type)

        return []

    def get_broker_ids(self):
        """
        获取broker_id
        :return:
        """
        lst_broker_id = []
        for broker in self.brokers:
            lst_broker_id.append(broker.id)
        return lst_broker_id



import datetime

from SapphireQuant.Config.TradingDay import TradingDayHelper
from SapphireQuant.Core.Enum import EnumBarType, EnumMarket, EnumQuoteEventType
from SapphireQuant.Simulate import ReplayData


class SimulateFutureQuoteChannel:
    """
    期货回测行情通道
    """

    def __init__(self, drive_bar_interval, drive_bar_type, begin_date, end_date, data_handler, debug_mode=False):
        self._on_start_event = []
        self._on_end_event = []
        self._on_bar_event = []
        self._on_tick_event = []
        self._on_day_begin_event = []
        self._on_day_end_event = []
        self._debug_mode = debug_mode

        self._drive_bar_interval = drive_bar_interval
        self._drive_bar_type = drive_bar_type
        self.start_simulate_date = TradingDayHelper.get_first_trading_day(begin_date)
        self.end_simulate_date = TradingDayHelper.get_last_trading_day(end_date)
        self._data_handler = data_handler

        self._drive_data = None
        self._replay_data_map = {}
        self.subscribe_futures = []
        self.trading_date = self.start_simulate_date
        self.tick_now = self.start_simulate_date
        self._drive_code = None

    def set_drive_code(self, drive_code):
        """

        :param drive_code:
        """
        self._drive_code = drive_code
        self._drive_data = ReplayData(drive_code)
        if self._drive_code not in self.subscribe_futures:
            self.subscribe_futures.append(self._drive_code)

    def subscribe(self, instruments):
        """
        订阅
        :param instruments:
        """
        for future_code in instruments:
            if len(future_code) == 0:
                continue

            if future_code not in self.subscribe_futures:
                self.subscribe_futures.append(future_code)

            if future_code == self._drive_code:
                continue

            if future_code not in self._replay_data_map.keys():
                self._replay_data_map[future_code] = ReplayData(future_code)

    def start(self):
        """
        开始播放
        """
        if self._drive_bar_type == EnumBarType.day:
            # 初始化数据
            for instrument_id in self.subscribe_futures:
                bar_series = self._data_handler.load_bar_series_by_date(
                    EnumMarket.期货, instrument_id, self._drive_bar_interval, self._drive_bar_type, self.start_simulate_date, self.end_simulate_date)
                if instrument_id == self._drive_code:
                    for bar in bar_series:
                        self._drive_data.add_data(bar.trading_date, [bar])
                else:
                    for bar in bar_series:
                        self._replay_data_map[instrument_id].add_data(bar.trading_date, [bar])

            self.on_start()
            lst_trading_days = TradingDayHelper.get_trading_days(self.start_simulate_date, self.end_simulate_date)
            for trading_date in lst_trading_days:
                self.trading_date = trading_date
                self._data_handler.init(trading_date)
                self.__replay_bar(trading_date)
            self.on_end()
        elif (self._drive_bar_type == EnumBarType.minute) or (self._drive_bar_type == EnumBarType.hour):
            self.on_start()
            lst_trading_days = TradingDayHelper.get_trading_days(self.start_simulate_date, self.end_simulate_date)
            for trading_date in lst_trading_days:
                self.trading_date = trading_date
                self._data_handler.init(trading_date)

                for instrument_id in self.subscribe_futures:
                    bar_series = self._data_handler.load_bar_series_by_date(
                        EnumMarket.期货, instrument_id, self._drive_bar_interval, self._drive_bar_type, trading_date, trading_date)
                    if instrument_id == self._drive_code:
                        self._drive_data.add_data(trading_date, bar_series)
                    else:
                        self._replay_data_map[instrument_id].add_data(trading_date, bar_series)

                self.__replay_bar(trading_date)
            self.on_end()
        elif self._drive_bar_type == EnumBarType.tick:
            self.on_start()
            lst_trading_days = TradingDayHelper.get_trading_days(self.start_simulate_date, self.end_simulate_date)
            for trading_date in lst_trading_days:
                self.trading_date = trading_date
                self._data_handler.init(trading_date)

                for instrument_id in self.subscribe_futures:
                    tick_series = self._data_handler.load_tick_series(EnumMarket.期货, instrument_id, trading_date)
                    if instrument_id == self._drive_code:
                        self._drive_data.add_data(trading_date, tick_series)
                    else:
                        self._replay_data_map[instrument_id].add_data(trading_date, tick_series)

                self.__replay_tick(trading_date)
            self.on_end()
        elif self._drive_bar_type == EnumBarType.night_am_pm:
            self.on_start()
            lst_trading_days = TradingDayHelper.get_trading_days(self.start_simulate_date, self.end_simulate_date)
            for trading_date in lst_trading_days:
                self.trading_date = trading_date
                self._data_handler.init(trading_date)

                for instrument_id in self.subscribe_futures:
                    bar_series = self._data_handler.load_night_am_pm_bar_series_by_date(
                        EnumMarket.期货, instrument_id, self._drive_bar_interval, self._drive_bar_type, trading_date, trading_date)
                    if instrument_id == self._drive_code:
                        self._drive_data.add_data(trading_date, bar_series)
                    else:
                        self._replay_data_map[instrument_id].add_data(trading_date, bar_series)

                self.__replay_bar(trading_date)
            self.on_end()

    def __replay_bar(self, trading_date):
        """
        播放bar
        :param trading_date:
        """
        self._drive_data.current_trading_date = trading_date
        for replay_data in self._replay_data_map.values():
            replay_data.current_trading_date = trading_date

        if trading_date in self._drive_data.data_block_map.keys():
            drive_data_series = self._drive_data.data_block_map[trading_date].data_list
        else:
            print('{0}@{1},缺失数据'.format(self._drive_code, self.trading_date))
            self.on_day_begin(self.trading_date)
            self.on_day_end()
            return

        trading_date = drive_data_series[0].trading_date
        self.on_day_begin(trading_date)
        for drive_data in drive_data_series:
            drive_data.instrument_id = self._drive_code
            self.tick_now = drive_data.end_time

            lst_replay_data = []
            for key in self._replay_data_map.keys():
                res = self._replay_data_map[key].replay_bar_to_time(self.tick_now)
                lst_replay_data = res[1]
                if res[0]:
                    for replay_data in lst_replay_data:
                        replay_data.instrument_id = key
            self.on_bar(drive_data, lst_replay_data)

        self.on_day_end()

    def __replay_tick(self, trading_date):
        """
        播放tick
        :param trading_date:
        """
        self._drive_data.current_trading_date = trading_date
        for replay_data in self._replay_data_map.values():
            replay_data.current_trading_date = trading_date

        if trading_date in self._drive_data.data_block_map.keys():
            drive_data_series = self._drive_data.data_block_map[trading_date].data_list
        else:
            print('{0}@{1},缺失数据'.format(self._drive_code, self.trading_date))
            self.on_day_begin(self.trading_date)
            self.on_day_end()
            return

        trading_date = drive_data_series[0].trading_date
        self.on_day_begin(trading_date)
        for drive_data in drive_data_series:
            drive_data.instrument_id = self._drive_code
            self.tick_now = drive_data.date_time

            lst_replay_data = []
            for key in self._replay_data_map.keys():
                res = self._replay_data_map[key].replay_tick_to_time(self.tick_now)
                lst_replay_data = res[1]
                if res[0]:
                    for replay_data in lst_replay_data:
                        replay_data.instrument_id = key
            self.on_tick(drive_data, lst_replay_data)

        self.on_day_end()

    def on_start(self):
        """
        回测开始
        """
        try:
            for event in self._on_start_event:
                event()
        except Exception as e:
            print('回测行情模块：[on_start]事件异常:{0}'.format(str(e)))
            if self._debug_mode:
                raise e

    def on_day_begin(self, trading_date):
        """
        交易日开始
        :param trading_date:
        """
        try:
            for event in self._on_day_begin_event:
                event(trading_date)
        except Exception as e:
            print('回测行情模块：[on_day_begin]事件异常:{0}'.format(str(e)))
            if self._debug_mode:
                raise e

    def on_day_end(self):
        """
        交易日结束
        """
        try:
            for event in self._on_day_end_event:
                event()
        except Exception as e:
            print('回测行情模块：[on_day_end]事件异常:{0}'.format(str(e)))
            if self._debug_mode:
                raise e

    def on_tick(self, drive_tick, lst_replay_tick):
        """
        on_tick触发
        :param lst_replay_tick:
        :param drive_tick:
        """
        try:
            for event in self._on_tick_event:
                event(drive_tick, lst_replay_tick)
        except Exception as e:
            print('回测行情模块：[on_tick]事件异常:{0}'.format(str(e)))
            if self._debug_mode:
                raise e

    def on_bar(self, drive_bar, lst_replay_bar):
        """
        OnBar触发
        :param lst_replay_bar: 其他Bar
        :param drive_bar:驱动Bar
        """
        try:
            for event in self._on_bar_event:
                event(drive_bar, lst_replay_bar)
        except Exception as e:
            print('回测行情模块：[on_bar]事件异常:{0}'.format(str(e)))
            if self._debug_mode:
                raise e

    def on_end(self):
        """
        回测结束
        """
        try:
            for event in self._on_end_event:
                event()
        except Exception as e:
            print('回测行情模块：[on_end]事件异常:{0}'.format(str(e)))
            if self._debug_mode:
                raise e

    def register_event(self, event, event_type):
        """
        注册事件
        :param event_type:
        :param event:EnumQuoteEventType
        """
        if event_type == EnumQuoteEventType.on_start:
            self._on_start_event.append(event)
        elif event_type == EnumQuoteEventType.on_end:
            self._on_end_event.append(event)
        elif event_type == EnumQuoteEventType.on_day_begin:
            self._on_day_begin_event.append(event)
        elif event_type == EnumQuoteEventType.on_day_end:
            self._on_day_end_event.append(event)
        elif event_type == EnumQuoteEventType.on_bar:
            self._on_bar_event.append(event)
        elif event_type == EnumQuoteEventType.on_tick:
            self._on_tick_event.append(event)


from .Future import Future
from .FutureAsset import FutureAsset
from .FutureExchange import FutureExchange
from .FutureOrder import FutureOrder
from .FuturePosition import FuturePosition
from .FuturePositionDetail import FuturePositionDetail
from .FutureProduct import FutureProduct
from .FutureTrade import FutureTrade

__all__ = [
    'Future',
    'FutureAsset',
    'FutureExchange',
    'FutureOrder',
    'FuturePosition',
    'FuturePositionDetail',
    'FutureProduct',
    'FutureTrade'
]


import datetime
import uuid

from SapphireQuant.Config.TradingDay.TradingDayHelper import TradingDayHelper
from SapphireQuant.Core.Enum import EnumPositionDirection, EnumOpenClose, EnumBuySell, EnumMarket, EnumHedgeFlag, EnumPositionType, EnumOrderStatus
from SapphireQuant.Core.Futures.FutureAsset import FutureAsset
from SapphireQuant.Core.Futures.FuturePosition import FuturePosition
from SapphireQuant.Core.Futures.FuturePositionDetail import FuturePositionDetail
from SapphireQuant.DataStorage.Cache.FutureData import FutureData
from SapphireQuant.DataStorage.Cache.PositionKey import PositionKey
from enum import Enum, unique
import pandas as pd


@unique
class EnumLiquidationType(Enum):
    """
    期货缓存清算方式
    """

    结算价 = 0,
    收盘价 = 1,


class FutureCache:
    """
    期货缓存
    和MQ，快期等对标，增加持仓明细的维护，平仓严格按照合约开仓的时间来进行先平老仓，后平新仓
    MQ的模式使用最新价进行清算
    快期等软件使用结算价进行清算
    """

    def __init__(self, transaction_rate, long_margin_rate, short_margin_rate, instrument_manager, data_handler, liquidation_type=EnumLiquidationType.结算价):
        self._trading_date = datetime.datetime.today()
        self._position_map = {}
        self._position_detail_map = {}
        self._order_map = {}
        self._trades = []
        self._asset = FutureAsset()
        self._liquidation_type = liquidation_type

        self._data_map = {}
        self._transaction_rate = transaction_rate
        self._long_margin_rate = long_margin_rate
        self._short_margin_rate = short_margin_rate
        self.instrument_manager = instrument_manager
        self.data_handler = data_handler

        self.all_assets = []
        self.all_orders = []
        self.all_trades = []
        self.all_positions = []
        self.all_position_details = []

        self.df_asset = None
        self.lst_asset = []

        self.df_order = None
        self.lst_order = []

        self.df_trade = None
        self.lst_trade = []

        self.df_position = None
        self.lst_position = []

        self.df_position_detail = None
        self.lst_position_detail = []

    def on_bar(self, bar):
        """
        on_bar维护缓存
        :param bar:
        """
        if bar.instrument_id not in self._data_map.keys():
            self._data_map[bar.instrument_id] = {}
        self._data_map[bar.instrument_id]['close'] = bar.close
        self._data_map[bar.instrument_id]['pre_close'] = bar.pre_close
        self._data_map[bar.instrument_id]['date_time'] = bar.end_time

    def on_tick(self, tick):
        """
        on_tick维护缓存
        :param tick:
        """
        if tick.instrument_id not in self._data_map.keys():
            self._data_map[tick.instrument_id] = {}
        self._data_map[tick.instrument_id]['close'] = tick.last_price
        self._data_map[tick.instrument_id]['pre_close'] = tick.pre_close_price
        self._data_map[tick.instrument_id]['date_time'] = tick.date_time

    def add_or_update_order(self, order):
        """
        添加或者更新order
        :param order:
        """
        if order.guid in self._order_map.keys():
            self._order_map[order.guid].update(order)
        else:
            self._order_map[order.guid] = order

    def add_trade(self, trade):
        """
        添加trade
        :param trade:
        """
        self._trades.append(trade)

    def calc_margin(self, position, volume_multiple):
        """
        计算保证金
        :param position:
        :param volume_multiple:
        :return:
        """
        if position.direction == EnumPositionDirection.多头:
            return position.volume * position.last_price * volume_multiple * self._long_margin_rate
        return position.volume * position.last_price * volume_multiple * self._short_margin_rate

    def on_order(self, order):
        """
        on_order触发
        :param order:
        """
        self.add_or_update_order(order)

    def on_order_canceled(self, order):
        """

        :param order:
        """
        self.add_or_update_order(order)

    def on_order_cancel_failed(self, order_guid, msg):
        """

        :param order_guid:
        :param msg:
        """
        pass

    def on_order_reject(self, order):
        """

        :param order:
        """
        self.add_or_update_order(order)

    def get_liquidation_info(self, instrument_id):
        """
        获取昨日收盘清算价格和当日收盘清算价格（pre_close & close） or (pre_settlement_price & settlement_price)
        :param instrument_id:
        :return:
        """
        last_tick = self.data_handler.load_last_tick(EnumMarket.期货, instrument_id, self._trading_date)
        if last_tick is not None:
            if self._liquidation_type == EnumLiquidationType.结算价:
                pre_liquidation_price = last_tick.pre_settlement_price
                # 当日结算价 下一日获取
                liquidation_price = self.data_handler.load_pre_settlement_price_by_tick(EnumMarket.期货, instrument_id, TradingDayHelper.get_next_trading_day(self._trading_date))
            else:
                pre_liquidation_price = last_tick.pre_close_price
                liquidation_price = last_tick.last_price
        else:
            pre_liquidation_price = 0.0
            liquidation_price = 0.0

        if pre_liquidation_price == 0.0:
            bar_series = self.data_handler.load_day_bar_series(EnumMarket.期货, instrument_id, TradingDayHelper.get_pre_trading_day(self._trading_date))
            if bar_series is not None and len(bar_series) > 0:
                print('未取到昨日清算价，根据昨日日线的收盘价来赋值结算价')
                pre_liquidation_price = bar_series[-1].close
            else:
                if instrument_id in self._data_map.keys():
                    if 'pre_close' in self._data_map[instrument_id].keys():
                        print('未取到昨日清算价，根据分钟线的昨收盘价来赋值结算价')
                        pre_liquidation_price = self._data_map[instrument_id]['pre_close']
                else:
                    print('未取到昨日清算价，用0来赋值结算价')
                    pre_liquidation_price = 0

        if liquidation_price == 0.0:
            print('未取到当日清算价，根据当日日线的收盘价来赋值结算价')
            bar_series = self.data_handler.load_day_bar_series(EnumMarket.期货, instrument_id, self._trading_date)
            if bar_series is not None and len(bar_series) > 0:
                liquidation_price = bar_series[-1].close
            else:
                if instrument_id in self._data_map.keys():
                    if 'close' in self._data_map[instrument_id].keys():
                        print('未取到昨日清算价，根据分钟线的收盘价来赋值结算价')
                        liquidation_price = self._data_map[instrument_id]['close']
                else:
                    print('未取到昨日清算价，用0来赋值结算价')
                    liquidation_price = 0
        return pre_liquidation_price, liquidation_price

    def on_trade(self, trade):
        """
        on_trade触发
        :param trade:
        :return:
        """
        future_info = self.instrument_manager[trade.instrument_id]
        multiplier = future_info.volume_multiple
        d_hand_fee = trade.trade_price * trade.trade_volume * multiplier * self._transaction_rate
        old_margin = 0

        liquidation_info = self.get_liquidation_info(trade.instrument_id)
        pre_liquidation_price = liquidation_info[0]

        # 持仓改变
        if trade.open_close == EnumOpenClose.开仓:
            if trade.direction == EnumBuySell.买入:
                position_direction = EnumPositionDirection.多头
            else:
                position_direction = EnumPositionDirection.空头
            position_key = PositionKey(trade.product_code, trade.strategy_id, trade.instrument_id, position_direction)

            # 添加position_detail
            position_detail = FuturePositionDetail()
            position_detail.product_code = trade.product_code
            position_detail.strategy_id = trade.strategy_id
            position_detail.trading_date = trade.trading_date
            position_detail.instrument_id = trade.instrument_id
            position_detail.instrument_name = trade.instrument_name
            position_detail.direction = position_direction
            position_detail.hedge_flag = EnumHedgeFlag.投机
            position_detail.exchange_id = trade.exchange_id
            position_detail.open_time = trade.trade_time
            position_detail.open_price = trade.trade_price
            position_detail.volume = trade.trade_volume
            position_detail.float_profit = 0.0
            position_detail.position_profit = 0.0
            position_detail.market_value = 0.0
            position_detail.position_type = EnumPositionType.今仓
            if position_key.to_string() in self._position_detail_map.keys():
                self._position_detail_map[position_key.to_string()].append(position_detail)
            else:
                self._position_detail_map[position_key.to_string()] = [position_detail]

            if position_key.to_string() in self._position_map.keys():
                position = self._position_map[position_key.to_string()]

                old_margin = self.calc_margin(position, multiplier)
                position.open_cost = (position.open_cost * position.volume + trade.trade_price * trade.trade_volume) / (position.volume + trade.trade_volume)
                position.td_open_trade += trade.trade_volume
                position.last_price = trade.trade_price
                position.transaction_cost += d_hand_fee
            else:
                # 添加持仓
                position = FuturePosition()
                position.product_code = trade.product_code
                position.strategy_id = trade.strategy_id
                position.trading_date = trade.trading_date
                position.exchange_id = trade.exchange_id
                position.instrument_id = trade.instrument_id
                position.instrument_name = trade.instrument_name
                position.open_cost = trade.trade_price
                position.volume = trade.trade_volume
                position.td_volume = trade.trade_volume
                position.transaction_cost = d_hand_fee
                position.direction = position_direction
                position.td_open_trade = trade.trade_volume

                self._position_map[position_key.to_string()] = position
        else:  # 平仓
            if trade.direction == EnumBuySell.买入:
                position_direction = EnumPositionDirection.空头
            else:
                position_direction = EnumPositionDirection.多头
            d_close_profit = 0

            # 反向持仓
            position_key = PositionKey(trade.product_code, trade.strategy_id, trade.instrument_id, position_direction)
            lst_position_detail = None
            position = None

            if position_key.to_string() in self._position_detail_map.keys():
                lst_position_detail = self._position_detail_map[position_key.to_string()]
                lst_position_detail.sort(key=lambda x: x.open_time)
                lst_left_position_detail = []

                sum_close_volume = 0
                volume_satisfied = False
                for position_detail in lst_position_detail:
                    if volume_satisfied:
                        lst_left_position_detail.append(position_detail)
                        continue

                    if (sum_close_volume + position_detail.volume) > trade.trade_volume:
                        close_volume = trade.trade_volume - sum_close_volume
                    else:
                        close_volume = position_detail.volume

                    sum_close_volume += position_detail.volume
                    if position_detail.position_type == EnumPositionType.昨仓:
                        if trade.direction == EnumBuySell.买入:
                            d_close_profit += close_volume * (trade.trade_price - pre_liquidation_price) * multiplier * -1  # 平仓是有盈亏的
                        else:
                            d_close_profit += close_volume * (trade.trade_price - pre_liquidation_price) * multiplier * 1  # 平仓是有盈亏的
                    elif position_detail.position_type == EnumPositionType.今仓:
                        if trade.direction == EnumBuySell.买入:
                            d_close_profit += close_volume * (trade.trade_price - position_detail.open_price) * multiplier * -1  # 平仓是有盈亏的
                        else:
                            d_close_profit += close_volume * (trade.trade_price - position_detail.open_price) * multiplier * 1  # 平仓是有盈亏的

                    if sum_close_volume == trade.trade_volume:
                        volume_satisfied = True
                    elif sum_close_volume > trade.trade_volume:
                        volume_satisfied = True
                        position_detail.volume = sum_close_volume - trade.trade_volume
                        lst_left_position_detail.append(position_detail)

                self._position_detail_map[position_key.to_string()] = lst_left_position_detail
            if lst_position_detail is None:
                return

            if position_key.to_string() in self._position_map.keys():
                position = self._position_map[position_key.to_string()]
            if position is None:
                return

            old_margin = self.calc_margin(position, multiplier)

            position.close_profit += d_close_profit
            position.td_close_trade += trade.trade_volume
            position.last_price = trade.trade_price
            position.transaction_cost += d_hand_fee

            self._asset.close_profit += d_close_profit
            self._asset.acc_close_profit += d_close_profit
            self._asset.cash += d_close_profit
            self._asset.fund_balance += d_close_profit

        if position_key.to_string() in self._position_detail_map.keys():
            lst_position_detail = self._position_detail_map[position_key.to_string()]
            volume = 0
            td_volume = 0
            yd_volume = 0
            sum_mv = 0
            for position_detail in lst_position_detail:
                volume += position_detail.volume
                if position_detail.position_type == EnumPositionType.今仓:
                    td_volume += position_detail.volume
                    sum_mv += position_detail.open_price * position_detail.volume
                elif position_detail.position_type == EnumPositionType.昨仓:
                    yd_volume += position_detail.volume
                    sum_mv += pre_liquidation_price * position_detail.volume

            position.volume = volume
            position.td_volume = td_volume
            position.yd_volume = yd_volume
            if position.volume != 0:
                position.open_cost = sum_mv / position.volume

        self._asset.cash -= d_hand_fee
        self._asset.fund_balance -= d_hand_fee
        self._asset.fee += d_hand_fee

        # 统一根据持仓价格计算保证金

        margin = self.calc_margin(position, multiplier)
        self._asset.margin += margin - old_margin
        self._asset.available_cash = self._asset.fund_balance - self._asset.margin

        self.add_trade(trade)

    def day_begin(self, trading_date, future_data):
        """
        当天开始，缓存初始化
        :param trading_date:
        :param future_data:
        """
        self._trading_date = trading_date
        self._position_map.clear()
        self._position_detail_map.clear()
        self._order_map.clear()
        self._trades.clear()
        self._asset = future_data.next_asset
        for position in future_data.next_positions:
            key = PositionKey(position.product_code, position.strategy_id, position.instrument_id, position.direction)
            self._position_map[key.to_string()] = position
        for position_detail in future_data.next_position_details:
            key = PositionKey(position_detail.product_code, position_detail.strategy_id, position_detail.instrument_id, position_detail.direction)
            if key.to_string() in self._position_detail_map.keys():
                self._position_detail_map[key.to_string()].append(position_detail)
            else:
                self._position_detail_map[key.to_string()] = [position_detail]

    def day_end(self):
        """
        当天结束，缓存清算
        :return:
        """
        next_trading_day = TradingDayHelper.get_next_trading_day(self._trading_date)
        self._asset.position_profit = 0
        d_deposit = 0
        d_market_value = 0
        position_details = []
        next_position_details = []
        for lst_position_detail in self._position_detail_map.values():
            for position_detail in lst_position_detail:
                position_details.append(position_detail)
                future_info = self.instrument_manager[position_detail.instrument_id]
                d_multiplier = future_info.volume_multiple

                liquidation_info = self.get_liquidation_info(position_detail.instrument_id)
                pre_liquidation_price = liquidation_info[0]
                liquidation_price = liquidation_info[1]

                position_detail.position_profit = 0.0
                if liquidation_price > 0:
                    position_profit = 0.0
                    if position_detail.position_type == EnumPositionType.今仓:
                        if position_detail.direction == EnumPositionDirection.多头:
                            position_profit += (liquidation_price - position_detail.open_price) * position_detail.volume * d_multiplier * 1
                        else:
                            position_profit += (liquidation_price - position_detail.open_price) * position_detail.volume * d_multiplier * -1
                    elif position_detail.position_type == EnumPositionType.昨仓:
                        if position_detail.direction == EnumPositionDirection.多头:
                            position_profit += (liquidation_price - pre_liquidation_price) * position_detail.volume * d_multiplier * 1
                        else:
                            position_profit += (liquidation_price - pre_liquidation_price) * position_detail.volume * d_multiplier * -1

                    self._asset.position_profit += position_profit
                    self._asset.fund_balance += position_profit
                    position_detail.position_profit = position_profit
                    if position_detail.direction == EnumPositionDirection.多头:
                        position_detail.market_value = liquidation_price * position_detail.volume * d_multiplier * 1
                    else:
                        position_detail.market_value = liquidation_price * position_detail.volume * d_multiplier * -1
                    d_market_value += position_detail.market_value

                if position_detail.volume > 0:
                    next_position_detail = position_detail.clone()
                    next_position_detail.guid = uuid.uuid1()
                    next_position_detail.position_type = EnumPositionType.昨仓
                    next_position_detail.trading_date = next_trading_day
                    next_position_detail.position_profit = 0
                    next_position_details.append(next_position_detail)

        next_positions = []
        for key in self._position_map.keys():
            position = self._position_map[key]
            lst_position_detail = []
            if key in self._position_detail_map.keys():
                lst_position_detail = self._position_detail_map[key]
            future_info = self.instrument_manager[position.instrument_id]
            d_multiplier = future_info.volume_multiple

            position.position_profit = 0.0

            liquidation_info = self.get_liquidation_info(position.instrument_id)
            liquidation_price = liquidation_info[1]

            position.last_price = liquidation_price
            position.market_value = 0
            for position_detail in lst_position_detail:
                position.market_value += position_detail.market_value
                position.position_profit += position_detail.position_profit

            d_deposit += self.calc_margin(position, d_multiplier)

            if position.volume > 0:
                next_position = position.clone()
                next_position.guid = uuid.uuid1()
                next_position.td_volume = 0
                next_position.yd_volume = position.volume
                next_position.close_volume = 0
                next_position.trading_date = next_trading_day
                next_position.td_close_trade = 0
                next_position.td_open_trade = 0
                next_position.close_profit = 0
                next_position.position_profit = 0
                next_position.num_holding_day = position.num_holding_day + 1
                next_positions.append(next_position)

        self._asset.margin = d_deposit
        self._asset.market_value = d_market_value
        self._asset.available_cash = self._asset.fund_balance - self._asset.margin

        next_asset = self._asset.clone()
        next_asset.guid = uuid.uuid1()
        next_asset.trading_date = next_trading_day
        next_asset.close_profit = 0
        next_asset.position_profit = 0
        next_asset.fee = 0
        next_asset.pre_fund_balance = self._asset.fund_balance

        next_future_data = FutureData()
        next_future_data.asset = self._asset
        next_future_data.positions = self._position_map.values()
        next_future_data.position_details = position_details
        next_future_data.orders = self._order_map.values()
        next_future_data.trades = []
        next_future_data.next_asset = next_asset
        next_future_data.next_positions = next_positions
        next_future_data.next_position_details = next_position_details

        for trade in self._trades:
            next_future_data.trades.append(trade)

        self._add_future_data(next_future_data)

        return next_future_data

    def on_end(self):
        """
        缓存结束，生成相应的DataFrame
        """
        self.df_asset = pd.DataFrame(self.lst_asset, columns=['产品', '策略', '日期', '昨日结存', '动态权益', '平仓盈亏', '持仓盈亏', '净利润', '可用资金', '持仓保证金', '手续费'])
        self.df_asset['累计盈亏'] = self.df_asset[['净利润']].cumsum(axis=0)
        self.df_asset['累计收益率'] = self.df_asset['累计盈亏'] / self.df_asset.iloc[0]['昨日结存']
        self.df_asset['累计收益率'] = self.df_asset['累计收益率'].apply(lambda x: format(x, '.2%'))

        self.df_order = pd.DataFrame(self.lst_order, columns=['产品', '策略', '交易日', '时间', '合约', '价格', '多空', '开平', '量', '状态'])
        self.df_trade = pd.DataFrame(self.lst_trade, columns=['产品', '策略', '交易日', '时间', '合约', '价格', '多空', '开平', '量'])
        self.df_position = pd.DataFrame(self.lst_position, columns=['产品', '策略', '交易日', '合约', '多空', '总持仓', '当前价', '持仓盈亏', '合约价值', '持有天数'])
        self.df_position_detail = pd.DataFrame(self.lst_position_detail, columns=['产品', '策略', '交易日', '合约', '多空', '总持仓', '开仓价', '开仓时间',
                                                                                  '持仓类型', '持仓盈亏', '浮动盈亏', '持仓市值'])

        self.df_trade['时间'] = self.df_trade['时间'].apply(lambda x: x.strftime('%H:%M:%S'))

    def _add_future_data(self, future_data):
        """
        新增期货清算数据
        :param future_data:
        """
        self.all_assets.append(future_data.asset)
        self.lst_asset.append([future_data.asset.product_code, future_data.asset.strategy_id,
                               future_data.asset.trading_date, future_data.asset.pre_fund_balance, future_data.asset.fund_balance,
                               future_data.asset.close_profit, future_data.asset.position_profit,
                               future_data.asset.close_profit + future_data.asset.position_profit,
                               future_data.asset.available_cash, future_data.asset.margin, future_data.asset.fee])

        for order in future_data.orders:
            self.all_orders.append(order)
            self.lst_order.append(
                [order.product_code, order.strategy_id, order.trading_date, order.order_time, order.instrument_id, order.order_price,
                 order.direction.chinese_name(), order.open_close.chinese_name(), order.order_volume,
                 order.order_status.chinese_name()])

        for trade in future_data.trades:
            self.all_trades.append(trade)
            self.lst_trade.append(
                [trade.product_code, trade.strategy_id, trade.trading_date, trade.trade_time, trade.instrument_id, trade.trade_price,
                 trade.direction.chinese_name(), trade.open_close.chinese_name(), trade.trade_volume])

        for position in future_data.positions:
            self.all_positions.append(position)
            self.lst_position.append([position.product_code, position.strategy_id, position.trading_date, position.instrument_id,
                                      position.direction.chinese_name(), position.volume, position.last_price,
                                      position.position_profit, position.market_value, position.num_holding_day])

        for position_detail in future_data.position_details:
            self.all_position_details.append(position_detail)
            self.lst_position_detail.append([position_detail.product_code, position_detail.strategy_id,
                                             position_detail.trading_date, position_detail.instrument_id,
                                             position_detail.direction.chinese_name(), position_detail.volume,
                                             position_detail.open_price, position_detail.open_time, position_detail.position_type.chinese_name(),
                                             position_detail.position_profit, position_detail.float_profit, position_detail.market_value])

    def get_position(self, product_code, strategy_id, instrument_id, direction):
        """
        获取缓存持仓
        :param product_code:
        :param strategy_id:
        :param instrument_id:
        :param direction:
        :return:
        """
        key = PositionKey(product_code, strategy_id, instrument_id, direction)
        key = key.to_string()
        if key in self._position_map.keys():
            return self._position_map[key]
        else:
            # 添加持仓
            position = FuturePosition()
            position.product_code = product_code
            position.strategy_id = strategy_id
            # position.ProjectId = trade.ProjectId
            position.trading_date = self._trading_date
            position.exchange_id = ''
            position.instrument_id = instrument_id
            position.instrument_name = ''
            position.open_cost = 0
            position.volume = 0
            position.td_volume = 0
            position.transaction_cost = 0
            position.direction = direction
            position.td_open_trade = 0

            self._position_map[key] = position
            return position

    def get_future_asset(self):
        """
        获取期货资金明细
        :return:
        """
        return self._asset

    def get_future_trades(self):
        """
        获取期货成交
        :return:
        """
        return self._trades

    def get_future_orders(self):
        """
        获取期货委托
        :return:
        """
        return list(self._order_map.values())

    def get_can_cancel_orders(self):
        """
        获取可撤委托
        :return:
        """
        order_series = []
        for order in self._order_map.values():
            if order.order_status != EnumOrderStatus.全部成交 and order.order_status != EnumOrderStatus.已撤:
                order_series.append(order)

        return order_series

    def get_future_positions(self):
        """
        获取期货持仓
        :return:
        """
        position_series = list(self._position_map.values())
        return position_series

    def get_future_order(self, guid):
        """
        获取期货委托
        :param guid:
        :return:
        """
        if guid in self._order_map.keys():
            return self._order_map[guid]
        return None





class FutureData:
    """
    期货清算数据
    """
    def __init__(self):
        self.asset = None
        self.positions = []
        self.position_details = []
        self.orders = []
        self.trades = []
        self.next_asset = None
        self.next_positions = []
        self.next_position_details = []



class PhysicalAccount:
    """
    物理账号
    """
    def __init__(self, user_id, password, broker_id, auth_code, channel_id='CTP'):
        self.user_id = user_id
        self.password = password
        self.encryption_password = ''
        self.channel_id = channel_id
        self.broker_id = broker_id
        self.auth_code = auth_code
        self.fund_unit = '{0}@{1}'.format(self.user_id, self.broker_id)

